﻿using System.Collections.Generic;
using System.Linq;

using Newtonsoft.Json;

namespace TestLibrary
{
    /// <summary>
    /// 필드 정보
    /// </summary>
    public class FieldInfo
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 제너레이터
        /// </summary>
        private IJSONClassGenerator generator;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 멤버명 - MemberName

        /// <summary>
        /// 멤버명
        /// </summary>
        public string MemberName { get; private set; }

        #endregion
        #region JSON 멤버명 - JSONMemberName

        /// <summary>
        /// JSON 멤버명
        /// </summary>
        public string JSONMemberName { get; private set; }

        #endregion
        #region 타입 - Type

        /// <summary>
        /// 타입
        /// </summary>
        public JSONType Type { get; private set; }

        #endregion
        #region 예제 리스트 - ExampleList

        /// <summary>
        /// 예제 리스트
        /// </summary>
        public IList<object> ExampleList { get; private set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - FieldInfo(generator, jsonMemberName, type, usePascalCase, exampleList)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="jsonMemberName">JSON 멤버명</param>
        /// <param name="type">타입</param>
        /// <param name="usePascalCase">파스칼 표기법 사용 여부</param>
        /// <param name="exampleList">예제 리스트</param>
        public FieldInfo(IJSONClassGenerator generator, string jsonMemberName, JSONType type, bool usePascalCase, IList<object> exampleList)
        {
            this.generator = generator;

            JSONMemberName = jsonMemberName;
            MemberName     = jsonMemberName;

            if(usePascalCase)
            {
                MemberName = JSONClassGenerator.ToTitleCase(MemberName);
            }

            Type        = type;
            ExampleList = exampleList;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성 코드 구하기 - GetGenerationCode(source)

        /// <summary>
        /// 생성 코드 구하기
        /// </summary>
        /// <param name="source">소스 문자열</param>
        /// <returns>생성 코드</returns>
        public string GetGenerationCode(string source)
        {
            FieldInfo fieldInfo = this;

            if(fieldInfo.Type.ValueType == JSONValueType.Array)
            {
                JSONType innermostType = fieldInfo.Type.GetInnermostType();

                return string.Format
                (
                    "({1})JsonClassHelper.ReadArray<{5}>(JsonClassHelper.GetJToken<JArray>({0}, \"{2}\"), JsonClassHelper.{3}, typeof({6}))",
                    source,
                    fieldInfo.Type.GetTypeName(),
                    fieldInfo.JSONMemberName,
                    innermostType.GetReaderName(),
                    -1,
                    innermostType.GetTypeName(),
                    fieldInfo.Type.GetTypeName()
                );
            }
            else if(fieldInfo.Type.ValueType == JSONValueType.Dictionary)
            {
                return string.Format
                (
                    "({1})JsonClassHelper.ReadDictionary<{2}>(JsonClassHelper.GetJToken<JObject>({0}, \"{3}\"))",
                    source,
                    fieldInfo.Type.GetTypeName(),
                    fieldInfo.Type.InternalType.GetTypeName(),
                    fieldInfo.JSONMemberName,
                    fieldInfo.Type.GetTypeName()
                );
            }
            else
            {
                return string.Format
                (
                    "JsonClassHelper.{1}(JsonClassHelper.GetJToken<{2}>({0}, \"{3}\"))",
                    source,
                    fieldInfo.Type.GetReaderName(),
                    fieldInfo.Type.GetJTokenType(),
                    fieldInfo.JSONMemberName
                );
            }
        }

        #endregion
        #region 예제 텍스트 구하기 - GetExampleText()

        /// <summary>
        /// 예제 텍스트 구하기
        /// </summary>
        /// <returns>예제 텍스트</returns>
        public string GetExampleText()
        {
            return string.Join(", ", ExampleList.Take(5).Select(x => JsonConvert.SerializeObject(x)).ToArray());
        }

        #endregion
    }
}