﻿using System.IO;

namespace TestLibrary
{
    /// <summary>
    /// 코드 작성자 인터페이스
    /// </summary>
    public interface ICodeWriter
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property

        #region 파일 확장자 - FileExtension

        /// <summary>
        /// 파일 확장자
        /// </summary>
        string FileExtension { get; }

        #endregion
        #region 디스플레이명 - DisplayName

        /// <summary>
        /// 디스플레이명
        /// </summary>
        string DisplayName { get; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method

        #region 타입명 구하기 - GetTypeName(type, generator)

        /// <summary>
        /// 타입명 구하기
        /// </summary>
        /// <param name="type">타입</param>
        /// <param name="generator">제너레이터</param>
        /// <returns>타입명</returns>
        string GetTypeName(JSONType type, IJSONClassGenerator generator);

        #endregion
        #region 클래스 쓰기 - WriteClass(generator, writer, type)

        /// <summary>
        /// 클래스 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="type">타입</param>
        void WriteClass(IJSONClassGenerator generator, TextWriter writer, JSONType type);

        #endregion
        #region 파일 시작 쓰기 - WriteFileStart(generator, writer)

        /// <summary>
        /// 파일 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        void WriteFileStart(IJSONClassGenerator generator, TextWriter writer);

        #endregion
        #region 파일 종료 쓰기 - WriteFileEnd(generator, writer)

        /// <summary>
        /// 파일 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        void WriteFileEnd(IJSONClassGenerator generator, TextWriter writer);

        #endregion
        #region 네임스페이스 시작 쓰기 - WriteNamespaceStart(generator, writer, root)

        /// <summary>
        /// 네임스페이스 시작 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        void WriteNamespaceStart(IJSONClassGenerator generator, TextWriter writer, bool root);

        #endregion
        #region 네임스페이스 종료 쓰기 - WriteNamespaceEnd(generator, writer, root)

        /// <summary>
        /// 네임스페이스 종료 쓰기
        /// </summary>
        /// <param name="generator">제너레이터</param>
        /// <param name="writer">작성자</param>
        /// <param name="root">루트 여부</param>
        void WriteNamespaceEnd(IJSONClassGenerator generator, TextWriter writer, bool root);

        #endregion
    }
}