﻿using System;
using System.Collections.Generic;
using System.Data.Entity.Design.PluralizationServices;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;

using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace TestLibrary
{
    /// <summary>
    /// JSON 클래스 제너레이터
    /// </summary>
    public class JSONClassGenerator : IJSONClassGenerator
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 파일 헤더 배열 - FileHeaderArray

        /// <summary>
        /// 파일 헤더 배열
        /// </summary>
        public static readonly string[] FileHeaderArray = new[]
        { 
            "Generated by JSON Class Generator", 
            "https://icodebroker.tistory.com"
        };

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 서비스
        /// </summary>
        private PluralizationService service = PluralizationService.CreateService(new CultureInfo("en-us"));

        /// <summary>
        /// 사용 여부
        /// </summary>
        private bool used = false;

        /// <summary>
        /// 명칭 해시 세트
        /// </summary>
        private HashSet<string> nameHashSet = new HashSet<string>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region (IJSONClassGenerator) 네임스페이스 - Namespace

        /// <summary>
        /// 네임스페이스
        /// </summary>
        public string Namespace { get; set; }

        #endregion
        #region (IJSONClassGenerator) 보조 네임스페이스 - SecondaryNamespace

        /// <summary>
        /// 보조 네임스페이스
        /// </summary>
        public string SecondaryNamespace { get; set; }

        #endregion
        #region (IJSONClassGenerator) 속성 사용 여부 - UseProperties

        /// <summary>
        /// 속성 사용 여부
        /// </summary>
        public bool UseProperties { get; set; }

        #endregion
        #region (IJSONClassGenerator) 내부 표시 여부 - InternalVisibility

        /// <summary>
        /// 내부 표시 여부
        /// </summary>
        public bool InternalVisibility { get; set; }

        #endregion
        #region (IJSONClassGenerator) 명시적 역직렬화 - ExplicitDeserialization

        /// <summary>
        /// 명시적 역직렬화
        /// </summary>
        public bool ExplicitDeserialization { get; set; }

        #endregion
        #region (IJSONClassGenerator) 헬퍼 클래스 없음 여부 - NoHelperClass

        /// <summary>
        /// 헬퍼 클래스 없음 여부
        /// </summary>
        public bool NoHelperClass { get; set; }

        #endregion
        #region (IJSONClassGenerator) 메인 클래스 - MainClass

        /// <summary>
        /// 메인 클래스
        /// </summary>
        public string MainClass { get; set; }

        #endregion
        #region (IJSONClassGenerator) 파스칼 표기법 사용 여부 - UsePascalCase

        /// <summary>
        /// 파스칼 표기법 사용 여부
        /// </summary>
        public bool UsePascalCase { get; set; }

        #endregion
        #region (IJSONClassGenerator) 중첩 클래스 사용 여부 - UseNestedClasses

        /// <summary>
        /// 중첩 클래스 사용 여부
        /// </summary>
        public bool UseNestedClasses { get; set; }

        #endregion
        #region (IJSONClassGenerator) 난독화 어트리뷰트 적용 여부 - ApplyObfuscationAttributes

        /// <summary>
        /// 난독화 어트리뷰트 적용 여부
        /// </summary>
        public bool ApplyObfuscationAttributes { get; set; }

        #endregion
        #region (IJSONClassGenerator) 단일 파일 여부 - SingleFile

        /// <summary>
        /// 단일 파일 여부
        /// </summary>
        public bool SingleFile { get; set; }

        #endregion
        #region (IJSONClassGenerator) 코드 작성자 - CodeWriter

        /// <summary>
        /// 코드 작성자
        /// </summary>
        public ICodeWriter CodeWriter { get; set; }

        #endregion
        #region (IJSONClassGenerator) 보조 클래스 보유 여부 - HasSecondaryClasses

        /// <summary>
        /// 보조 클래스 보유 여부
        /// </summary>
        public bool HasSecondaryClasses
        {
            get
            {
                return TypeList.Count > 1;
            }
        }

        #endregion
        #region (IJSONClassGenerator) NULL 가능 값 항상 사용 여부 - AlwaysUseNullableValues

        /// <summary>
        /// NULL 가능 값 항상 사용 여부
        /// </summary>
        public bool AlwaysUseNullableValues { get; set; }

        #endregion
        #region (IJSONClassGenerator) 네임스페이스 사용 여부 - UseNamespaces

        /// <summary>
        /// 네임스페이스 사용 여부
        /// </summary>
        public bool UseNamespaces
        {
            get
            {
                return Namespace != null;
            }
        }

        #endregion
        #region (IJSONClassGenerator) 문서 내 예제 여부 - ExamplesInDocumentation

        /// <summary>
        /// 문서 내 예제 여부
        /// </summary>
        public bool ExamplesInDocumentation { get; set; }

        #endregion

        #region 예제 - Example

        /// <summary>
        /// 예제
        /// </summary>
        public string Example { get; set; }

        #endregion
        #region 타겟 디렉토리 경로 - TargetDirectoryPath

        /// <summary>
        /// 타겟 디렉토리 경로
        /// </summary>
        public string TargetDirectoryPath { get; set; }

        #endregion
        #region 작성자 - Writer

        /// <summary>
        /// 작성자
        /// </summary>
        public TextWriter Writer { get; set; }

        #endregion
        #region 타입 리스트 - TypeList

        /// <summary>
        /// 타입 리스트
        /// </summary>
        public IList<JSONType> TypeList { get; private set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 클래스 생성하기 - GenerateClasses()

        /// <summary>
        /// 클래스 생성하기
        /// </summary>
        public void GenerateClasses()
        {
            if(CodeWriter == null)
            {
                CodeWriter = new CSharpCodeWriter();
            }

            if(ExplicitDeserialization && !(CodeWriter is CSharpCodeWriter))
            {
                throw new ArgumentException("Explicit deserialization is obsolete and is only supported by the C# provider.");
            }

            if(this.used)
            {
                throw new InvalidOperationException("This instance of JsonClassGenerator has already been used. Please create a new instance.");
            }

            this.used = true;

            bool writeToDisk = TargetDirectoryPath != null;

            if(writeToDisk && !Directory.Exists(TargetDirectoryPath))
            {
                Directory.CreateDirectory(TargetDirectoryPath);
            }

            JObject[] exampleArray;

            string example = Example.StartsWith("HTTP/") ? Example.Substring(Example.IndexOf("\r\n\r\n")) : Example;

            using(StringReader stringReader = new StringReader(example))
            {
                using(JsonTextReader jsonTextReader = new JsonTextReader(stringReader))
                {
                    JToken token = JToken.ReadFrom(jsonTextReader);

                    if(token is JArray)
                    {
                        exampleArray = ((JArray)token).Cast<JObject>().ToArray();
                    }
                    else if(token is JObject)
                    {
                        exampleArray = new[] { (JObject)token };
                    }
                    else
                    {
                        throw new Exception("Sample JSON must be either a JSON array, or a JSON object.");
                    }
                }
            }

            TypeList = new List<JSONType>();

            this.nameHashSet.Add(MainClass);

            JSONType rootType = new JSONType(this, exampleArray[0]);

            rootType.IsRoot = true;

            rootType.AssignName(MainClass);

            GenerateClass(exampleArray, rootType);

            if(writeToDisk)
            {
                string parentFolder = Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);

                if(writeToDisk && !NoHelperClass && ExplicitDeserialization)
                {
                    File.WriteAllBytes(Path.Combine(TargetDirectoryPath, "JSONClassHelper.cs"), Properties.Resources.JSONClassHelper);
                }

                if(SingleFile)
                {
                    WriteClassesToFile(Path.Combine(TargetDirectoryPath, MainClass + CodeWriter.FileExtension), TypeList);
                }
                else
                {
                    foreach(JSONType type in TypeList)
                    {
                        string directoryPath = TargetDirectoryPath;

                        if(!UseNestedClasses && !type.IsRoot && SecondaryNamespace != null)
                        {
                            string temporaryNamespace = SecondaryNamespace;

                            if(temporaryNamespace.StartsWith(Namespace + "."))
                            {
                                temporaryNamespace = temporaryNamespace.Substring(Namespace.Length + 1);
                            }

                            directoryPath = Path.Combine(directoryPath, temporaryNamespace);

                            Directory.CreateDirectory(directoryPath);
                        }

                        WriteClassesToFile
                        (
                            Path.Combine(directoryPath, (UseNestedClasses && !type.IsRoot ? MainClass + "." : string.Empty) + type.AssignedName + CodeWriter.FileExtension),
                            new[] { type }
                        );
                    }
                }
            }
            else if(Writer != null)
            {
                WriteClassesToFile(Writer, TypeList);
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region 제목 케이스 구하기 - ToTitleCase(source)

        /// <summary>
        /// 제목 케이스 구하기
        /// </summary>
        /// <param name="source">소스 문자열</param>
        /// <returns>제목 케이스</returns>
        internal static string ToTitleCase(string source)
        {
            StringBuilder stringBuilder = new StringBuilder(source.Length);

            bool flag = true;

            for(int i = 0; i < source.Length; i++)
            {
                char character = source[i];

                if(char.IsLetterOrDigit(character))
                {
                    stringBuilder.Append(flag ? char.ToUpper(character) : character);

                    flag = false;
                }
                else
                {
                    flag = true;
                }
            }

            return stringBuilder.ToString();
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 유일한 클래스명 생성하기 - CreateUniqueClassName(name)

        /// <summary>
        /// 유일한 클래스명 생성하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>유일한 클래스명</returns>
        private string CreateUniqueClassName(string name)
        {
            name = ToTitleCase(name);

            string finalName = name;

            int i = 2;

            while(this.nameHashSet.Any(x => x.Equals(finalName, StringComparison.OrdinalIgnoreCase)))
            {
                finalName = name + i.ToString();

                i++;
            }

            this.nameHashSet.Add(finalName);

            return finalName;
        }

        #endregion
        #region 복수의 값에서 유일한 클래스명 생성하기 - CreateUniqueClassNameFromPlural(plural)

        /// <summary>
        /// 복수의 문자열에서 유일한 클래스명 생성하기
        /// </summary>
        /// <param name="plural">복수의 문자열</param>
        /// <returns>유일한 클래스명</returns>
        private string CreateUniqueClassNameFromPlural(string plural)
        {
            plural = ToTitleCase(plural);

            return CreateUniqueClassName(this.service.Singularize(plural));
        }

        #endregion
        #region 클래스 생성하기 - GenerateClass(exampleArray, type)

        /// <summary>
        /// 클래스 생성하기
        /// </summary>
        /// <param name="exampleArray">예제 배열</param>
        /// <param name="type">타입</param>
        private void GenerateClass(JObject[] exampleArray, JSONType type)
        {
            Dictionary<string, JSONType> jsonFieldDictionary = new Dictionary<string, JSONType>();

            Dictionary<string, IList<object>> fieldExampleDictionary = new Dictionary<string, IList<object>>();

            bool first = true;

            foreach(JObject jObject in exampleArray)
            {
                foreach(JProperty jProperty in jObject.Properties())
                {
                    JSONType fieldType;

                    JSONType currentType = new JSONType(this, jProperty.Value);

                    string propertyName = jProperty.Name;

                    if(jsonFieldDictionary.TryGetValue(propertyName, out fieldType))
                    {
                        JSONType commonType = fieldType.GetCommonType(currentType);

                        jsonFieldDictionary[propertyName] = commonType;
                    }
                    else
                    {
                        JSONType commonType = currentType;

                        if(first)
                        {
                            commonType = commonType.MaybeMakeNullable(this);
                        }
                        else
                        {
                            commonType = commonType.GetCommonType(JSONType.GetNull(this));
                        }

                        jsonFieldDictionary.Add(propertyName, commonType);

                        fieldExampleDictionary[propertyName] = new List<object>();
                    }

                    IList<object> fieldExample = fieldExampleDictionary[propertyName];

                    JToken token = jProperty.Value;

                    if(token.Type == JTokenType.Null || token.Type == JTokenType.Undefined)
                    {
                        if(!fieldExample.Contains(null))
                        {
                            fieldExample.Insert(0, null);
                        }
                    }
                    else
                    {
                        object value = token.Type == JTokenType.Array || token.Type == JTokenType.Object ? token : token.Value<object>();

                        if(!fieldExample.Any(x => value.Equals(x)))
                        {
                            fieldExample.Add(value);
                        }
                    }
                }

                first = false;
            }

            if(UseNestedClasses)
            {
                foreach(KeyValuePair<string, JSONType> keyValuePair in jsonFieldDictionary)
                {
                    this.nameHashSet.Add(keyValuePair.Key.ToLower());
                }
            }

            foreach(KeyValuePair<string, JSONType> keyValuePair in jsonFieldDictionary)
            {
                JSONType fieldType = keyValuePair.Value;

                if(fieldType.ValueType == JSONValueType.Object)
                {
                    List<JObject> subsidaryExampleList = new List<JObject>(exampleArray.Length);

                    foreach(JObject jObject in exampleArray)
                    {
                        JToken value;

                        if(jObject.TryGetValue(keyValuePair.Key, out value))
                        {
                            if(value.Type == JTokenType.Object)
                            {
                                subsidaryExampleList.Add((JObject)value);
                            }
                        }
                    }

                    fieldType.AssignName(CreateUniqueClassName(keyValuePair.Key));

                    GenerateClass(subsidaryExampleList.ToArray(), fieldType);
                }

                if(fieldType.InternalType != null && fieldType.InternalType.ValueType == JSONValueType.Object)
                {
                    List<JObject> subsidaryExampleList = new List<JObject>(exampleArray.Length);

                    foreach(JObject jObject in exampleArray)
                    {
                        JToken value;

                        if(jObject.TryGetValue(keyValuePair.Key, out value))
                        {
                            if(value.Type == JTokenType.Array)
                            {
                                foreach(JToken item in (JArray)value)
                                {
                                    if(!(item is JObject))
                                    {
                                        throw new NotSupportedException("Arrays of non-objects are not supported yet.");
                                    }

                                    subsidaryExampleList.Add((JObject)item);
                                }
                            }
                            else if(value.Type == JTokenType.Object)
                            {
                                foreach(KeyValuePair<string, JToken> item in (JObject)value)
                                {
                                    if(!(item.Value is JObject))
                                    {
                                        throw new NotSupportedException("Arrays of non-objects are not supported yet.");
                                    }

                                    subsidaryExampleList.Add((JObject)item.Value);
                                }
                            }
                        }
                    }

                    keyValuePair.Value.InternalType.AssignName(CreateUniqueClassNameFromPlural(keyValuePair.Key));

                    GenerateClass(subsidaryExampleList.ToArray(), keyValuePair.Value.InternalType);
                }
            }

            type.FieldInfoList = jsonFieldDictionary.Select(x => new FieldInfo(this, x.Key, x.Value, UsePascalCase, fieldExampleDictionary[x.Key])).ToArray();

            TypeList.Add(type);
        }

        #endregion
        #region 파일에 클래스 쓰기 - WriteClassesToFile(writer, typeEnumerable)

        /// <summary>
        /// 파일에 클래스 쓰기
        /// </summary>
        /// <param name="writer">작성자</param>
        /// <param name="typeEnumerable">타입 열거 가능형</param>
        private void WriteClassesToFile(TextWriter writer, IEnumerable<JSONType> typeEnumerable)
        {
            bool inNamespace   = false;
            bool rootNamespace = false;

            CodeWriter.WriteFileStart(this, writer);

            foreach(JSONType type in typeEnumerable)
            {
                if(UseNamespaces && inNamespace && rootNamespace != type.IsRoot && SecondaryNamespace != null)
                {
                    CodeWriter.WriteNamespaceEnd(this, writer, rootNamespace);
                    
                    inNamespace = false;
                }

                if(UseNamespaces && !inNamespace)
                {
                    CodeWriter.WriteNamespaceStart(this, writer, type.IsRoot);
                    
                    inNamespace = true; rootNamespace = type.IsRoot;
                }

                CodeWriter.WriteClass(this, writer, type);
            }

            if(UseNamespaces && inNamespace)
            {
                CodeWriter.WriteNamespaceEnd(this, writer, rootNamespace);
            }

            CodeWriter.WriteFileEnd(this, writer);
        }

        #endregion
        #region 파일에 클래스 쓰기 - WriteClassesToFile(filePath, typeEnumerable)

        /// <summary>
        /// 파일에 클래스 쓰기
        /// </summary>
        /// <param name="filePath">파일 경로</param>
        /// <param name="typeEnumerable">타입 열거 가능형</param>
        private void WriteClassesToFile(string filePath, IEnumerable<JSONType> typeEnumerable)
        {
            using(StreamWriter writer = new StreamWriter(filePath, false, Encoding.UTF8))
            {
                WriteClassesToFile(writer, typeEnumerable);
            }
        }

        #endregion
    }
}