﻿using System;
using System.IO;
using System.Linq;
using System.Windows.Forms;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    /// 메인 폼
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 코드 작성자 배열
        /// </summary>
        private readonly static ICodeWriter[] _codeWriterArray = new ICodeWriter[]
        {
            new CSharpCodeWriter(),
            new VisualBasicCodeWriter(),
            new TypeScriptCodeWriter()
        };

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 마지막 생성 문자열
        /// </summary>
        private string lastGeneratedString;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainForm()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            Load                                                 += Form_Load;
            FormClosing                                          += Form_FormClosing;
            this.namespaceTextBox.TextChanged                    += namespaceTextBox_TextChanged;
            this.mainClassTextBox.TextChanged                    += mainClassTextBox_TextChanged;
            this.targetDirectoryPathButton.Click                 += targetDirectoryPathButton_Click;
            this.languageComboBox.SelectedIndexChanged           += languageComboBox_SelectedIndexChanged;
            this.useNestedClassesRadioButton.CheckedChanged      += useNestedClassesRadioButton_CheckedChanged;
            this.useSameNamespaceRadioButton.CheckedChanged      += useSameNamespaceRadioButton_CheckedChanged;
            this.useDifferentNamespaceRadioButton.CheckedChanged += useDifferentNamespaceRadioButton_CheckedChanged;
            this.explicitDeserializationCheckBox.CheckedChanged  += explicitDeserializationCheckBox_CheckedChanged;
            this.jsonTextBox.KeyDown                             += jsonTextBox_KeyDown;
            this.pasteButton.Click                               += pasteButton_Click;
            this.generateButton.Click                            += generateButton_Click;
            this.closeButton.Click                               += closeButton_Click;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected
        //////////////////////////////////////////////////////////////////////////////// Function

        #region 키 DOWN 처리하기 - OnKeyDown(e)

        /// <summary>
        /// 키 DOWN 처리하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if(e.KeyCode == Keys.F9)
            {
                e.Handled = true;

                Paste();
            }

            base.OnKeyDown(e);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 폼 로드시 처리하기 - Form_Load(sender, e)

        /// <summary>
        /// 폼 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_Load(object sender, EventArgs e)
        {
            Properties.Settings settings = Properties.Settings.Default;

            (settings.UseProperties      ? this.propertyRadioButton : this.fieldRadioButton ).Checked = true;
            (settings.InternalVisibility ? this.internalRadioButton : this.publicRadioButton).Checked = true;

            this.secondaryNamespaceTextBox.Text = settings.SecondaryNamespace;

            if(settings.NamespaceStrategy == 0)
            {
                this.useNestedClassesRadioButton.Checked = true;
            }
            else if(settings.NamespaceStrategy == 1)
            {
                this.useSameNamespaceRadioButton.Checked = true;
            }
            else
            {
                this.useDifferentNamespaceRadioButton.Checked = true;
            }

            this.languageComboBox.Items.AddRange(_codeWriterArray);

            int languageIndex = _codeWriterArray.ToList().FindIndex(x => x.GetType().Name == settings.Language);

            this.languageComboBox.SelectedIndex = languageIndex != -1 ? languageIndex : 0;

            this.singleFileCheckBox.Checked = settings.SingleFile;

            this.examplesInDocumentationCheckBox.Checked = settings.DocumentationExamples;

            UpdateStatus();
        }

        #endregion
        #region 폼을 닫을 경우 처리하기 - Form_FormClosing(sender, e)

        /// <summary>
        /// 폼을 닫을 경우 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Form_FormClosing(object sender, FormClosingEventArgs e)
        {
            Properties.Settings settings = Properties.Settings.Default;

            settings.UseProperties      = this.propertyRadioButton.Checked;
            settings.InternalVisibility = this.internalRadioButton.Checked;
            settings.SecondaryNamespace = this.secondaryNamespaceTextBox.Text;

            if(this.useNestedClassesRadioButton.Checked)
            {
                settings.NamespaceStrategy = 0;
            }
            else if(this.useSameNamespaceRadioButton.Checked)
            {
                settings.NamespaceStrategy = 1;
            }
            else
            {
                settings.NamespaceStrategy = 2;
            }

            if(!settings.UseSeparateNamespace)
            {
                settings.SecondaryNamespace = string.Empty;
            }

            settings.Language              = this.languageComboBox.SelectedItem.GetType().Name;
            settings.SingleFile            = this.singleFileCheckBox.Checked;
            settings.DocumentationExamples = this.examplesInDocumentationCheckBox.Checked;

            settings.Save();
        }

        #endregion
        #region 네임스페이스 텍스트 박스 텍스트 변경시 처리하기 - namespaceTextBox_TextChanged(sender, e)

        /// <summary>
        /// 네임스페이스 텍스트 박스 텍스트 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void namespaceTextBox_TextChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region 메인 클래스 텍스트 박스 텍스트 변경시 처리하기 - mainClassTextBox_TextChanged(sender, e)

        /// <summary>
        /// 메인 클래스 텍스트 박스 텍스트 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void mainClassTextBox_TextChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region 타겟 디렉토리 경로 버튼 클릭시 처리하기 - targetDirectoryPathButton_Click(sender, e)

        /// <summary>
        /// 타겟 디렉토리 경로 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void targetDirectoryPathButton_Click(object sender, EventArgs e)
        {
            using(FolderBrowserDialog dialog = new FolderBrowserDialog())
            {
                dialog.ShowNewFolderButton = true;
                dialog.SelectedPath        = this.targetDirectoryPathTextBox.Text;
                dialog.Description         = "생성 파일을 저장하기 위한 디렉토리를 선택해 주시기 바랍니다.";

                if(dialog.ShowDialog(this) == DialogResult.OK)
                {
                    this.targetDirectoryPathTextBox.Text = dialog.SelectedPath;
                }
            }
        }

        #endregion
        #region 언어 콤보 박스 선택 인덱스 변경시 처리하기 - languageComboBox_SelectedIndexChanged(sender, e)

        /// <summary>
        /// 언어 콤보 박스 선택 인덱스 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void languageComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region 중첩 클래스 사용 라디오 버튼 체크 변경시 처리하기 - useNestedClassesRadioButton_CheckedChanged(sender, e)

        /// <summary>
        /// 중첩 클래스 사용 라디오 버튼 체크 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void useNestedClassesRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region 동일 네임스페이스 라디오 버튼 체크 변경시 처리하기 - useSameNamespaceRadioButton_CheckedChanged(sender, e)

        /// <summary>
        /// 동일 네임스페이스 라디오 버튼 체크 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void useSameNamespaceRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region 다른 네임스페이스 라디오 버튼 체크 변경시 처리하기 - useDifferentNamespaceRadioButton_CheckedChanged(sender, e)

        /// <summary>
        /// 다른 네임스페이스 라디오 버튼 체크 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void useDifferentNamespaceRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region 명시적 역직렬화 사용 체크 박스 체크 변경시 처리하기 - explicitDeserializationCheckBox_CheckedChanged(sender, e)

        /// <summary>
        /// 명시적 역직렬화 사용 체크 박스 체크 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void explicitDeserializationCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            UpdateStatus();
        }

        #endregion
        #region JSON 텍스트 박스 키 DOWN 처리하기 - jsonTextBox_KeyDown(sender, e)

        /// <summary>
        /// JSON 텍스트 박스 키 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void jsonTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if(e.Control && e.KeyCode == Keys.A)
            {
                this.jsonTextBox.SelectionStart  = 0;
                this.jsonTextBox.SelectionLength = this.jsonTextBox.TextLength;

                e.Handled = true;
            }
        }

        #endregion
        #region 붙여넣기 버튼 클릭시 처리하기 - pasteButton_Click(sender, e)

        /// <summary>
        /// 붙여넣기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pasteButton_Click(object sender, EventArgs e)
        {
            Paste();
        }

        #endregion
        #region 생성 버튼 클릭시 처리하기 - generateButton_Click(sender, e)

        /// <summary>
        /// 생성 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void generateButton_Click(object sender, EventArgs e)
        {
            if(this.namespaceTextBox.Text == string.Empty)
            {
                MessageBox.Show
                (
                    this,
                    "네임스페이스를 지정해 주시기 바랍니다.",
                    "INFORMATION",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Information
                );

                return;
            }

            if(this.targetDirectoryPathTextBox.Text == string.Empty)
            {
                MessageBox.Show
                (
                    this,
                    "타겟 디렉토리 경로를 지정해 주시기 바랍니다.",
                    "INFORMATION",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Information
                );

                return;
            }

            JSONClassGenerator generator = GetGenerator();

            if(generator == null)
            {
                return;
            }

            try
            {
                generator.GenerateClasses();
            }
            catch(Exception exception)
            {
                MessageBox.Show
                (
                    this,
                    "코드를 생성할 수 없습니다 : " + exception.Message,
                    "ERROR",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error
                );
            }
        }

        #endregion
        #region 닫기 버튼 클릭시 처리하기 - closeButton_Click(sender, e)

        /// <summary>
        /// 닫기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void closeButton_Click(object sender, EventArgs e)
        {
            Close();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 상태 업데이트하기 - UpdateStatus()

        /// <summary>
        /// 상태 업데이트하기
        /// </summary>
        private void UpdateStatus()
        {
            if(this.secondaryNamespaceTextBox.Text.Contains("JsonTypes") || this.secondaryNamespaceTextBox.Text == string.Empty)
            {
                this.secondaryNamespaceTextBox.Text = this.namespaceTextBox.Text == string.Empty ? string.Empty : this.namespaceTextBox.Text + "." + "JsonTypes";
            }

            this.secondaryNamespaceTextBox.Enabled = this.useDifferentNamespaceRadioButton.Checked;

            ICodeWriter writer = this.languageComboBox.SelectedItem as ICodeWriter;

            this.pascalCaseCheckBox.Enabled = !(writer is TypeScriptCodeWriter);

            this.explicitDeserializationCheckBox.Enabled = writer is CSharpCodeWriter;

            this.noHelperClassCheckBox.Enabled = this.explicitDeserializationCheckBox.Enabled && this.explicitDeserializationCheckBox.Checked;
        }

        #endregion
        #region 제너레이터 구하기 - GetGenerator()

        /// <summary>
        /// 제너레이터 구하기
        /// </summary>
        /// <returns>제너레이터</returns>
        private JSONClassGenerator GetGenerator()
        {
            if(this.jsonTextBox.Text == string.Empty)
            {
                MessageBox.Show
                (
                    this,
                    "JSON을 입력해 주시기 바랍니다.",
                    "INFORMATION",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Information
                );

                this.jsonTextBox.Focus();

                return null;
            }

            if(this.mainClassTextBox.Text == string.Empty)
            {
                MessageBox.Show
                (
                    this,
                    "메인 클래스명을 입력해 주시기 바랍니다.",
                    "INFORMATION",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Information
                );

                return null;
            }

            JSONClassGenerator generator = new JSONClassGenerator();

            generator.Example                    = this.jsonTextBox.Text;
            generator.InternalVisibility         = this.internalRadioButton.Checked;
            generator.CodeWriter                 = this.languageComboBox.SelectedItem as ICodeWriter;
            generator.ExplicitDeserialization    = this.explicitDeserializationCheckBox.Checked && generator.CodeWriter is CSharpCodeWriter;
            generator.Namespace                  = string.IsNullOrEmpty(this.namespaceTextBox.Text) ? null : this.namespaceTextBox.Text;
            generator.NoHelperClass              = this.noHelperClassCheckBox.Checked;
            generator.SecondaryNamespace         = this.useDifferentNamespaceRadioButton.Checked &&
                                                       !string.IsNullOrEmpty(this.secondaryNamespaceTextBox.Text) ? this.secondaryNamespaceTextBox.Text : null;
            generator.TargetDirectoryPath        = this.targetDirectoryPathTextBox.Text;
            generator.UseProperties              = this.propertyRadioButton.Checked;
            generator.MainClass                  = this.mainClassTextBox.Text;
            generator.UsePascalCase              = this.pascalCaseCheckBox.Checked;
            generator.UseNestedClasses           = this.useNestedClassesRadioButton.Checked;
            generator.ApplyObfuscationAttributes = this.applyObfuscationAttributeCheckBox.Checked;
            generator.SingleFile                 = this.singleFileCheckBox.Checked;
            generator.ExamplesInDocumentation    = this.examplesInDocumentationCheckBox.Checked;

            return generator;
        }

        #endregion
        #region 붙여넣기 - Paste()

        /// <summary>
        /// 붙여넣기
        /// </summary>
        private void Paste()
        {
            string jsonClipboard = Clipboard.GetText(TextDataFormat.UnicodeText | TextDataFormat.Text);

            if(jsonClipboard != this.lastGeneratedString)
            {
                string jsonClipboardTrimmed = jsonClipboard.Trim();

                string jsonTextboxTrimmed = this.jsonTextBox.Text.Trim();

                if((jsonClipboardTrimmed.StartsWith("{") || jsonClipboardTrimmed.StartsWith("[")) || !(jsonTextboxTrimmed.StartsWith("{") || jsonTextboxTrimmed.StartsWith("[")))
                {
                    this.jsonTextBox.Text = jsonClipboard;
                }
            }

            JSONClassGenerator generator = GetGenerator();

            if(generator == null)
            {
                return;
            }

            try
            {
                generator.TargetDirectoryPath = null;
                generator.SingleFile          = true;

                using(StringWriter writer = new StringWriter())
                {
                    generator.Writer = writer;

                    generator.GenerateClasses();

                    writer.Flush();

                    this.lastGeneratedString = writer.ToString();

                    Clipboard.SetText(this.lastGeneratedString);
                }
            }
            catch(Exception exception)
            {
                MessageBox.Show
                (
                    this,
                    "코드를 생성할 수 없습니다 : " + exception.Message,
                    "ERROR",
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error
                );
            }
        }

        #endregion
    }
}